-- | Calculus module (evaluator).
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module Calculus
    ( compute
    , eval1
    , eval
    , fulleval
    , safeeval
    , safefulleval
    , smarteval
    , smartfulleval
    ) where

import Terms
import Sidekicks
import Text.ParserCombinators.Parsec.Error ( ParseError )


-- | Single-step evaluation.
eval1 :: Term -> Term

-- boolean
eval1 TmTrue                          = TmStop
eval1 TmFalse                         = TmStop
eval1 (TmIf TmTrue t1 _)              = t1
eval1 (TmIf TmFalse _ t2)             = t2
eval1 (TmIf t t1 t2)                  = let t' = eval1 t in
                                        if t' == TmStop then TmStop else TmIf t' t1 t2

-- arithmetic
eval1 TmZero                          = TmStop
eval1 (TmSucc t)                      = let t' = eval1 t in
                                        if t' == TmStop then TmStop else TmSucc t'
eval1 (TmPred TmZero)                 = TmZero
eval1 (TmPred t)                      = case t of
                                        (TmSucc nv) -> if isNatValue nv then nv else TmStop
                                        _           -> (TmPred t') where t' = eval1 t
eval1 (TmIsZero TmZero)               = TmTrue
eval1 (TmIsZero (TmSucc TmZero))      = TmFalse  
eval1 (TmIsZero (TmSucc t))           = let t' = eval1 t in
                                        if t' == TmStop then TmFalse else TmIsZero (TmSucc t')

-- lambda calculus
eval1 (TmAbs x (TmError e))           = TmError e
eval1 (TmAbs x t)                     = let t' = eval1 t in
                                        if t' == TmStop then TmStop else TmAbs x t'
eval1 ((TmAbs x t) `TmApp` t2)        = (x |-> t2) t
eval1 ((TmVar x)   `TmApp` (TmVar y)) = TmStop
eval1 ((TmVar x)   `TmApp` t2)        = let t2' = eval1 t2 in
                                        if t2' == TmStop then TmStop else ((TmVar x) `TmApp` t2')
eval1 (t1          `TmApp` t2)        = let t1' = eval1 t1 in
                                        if t1' == TmStop then TmStop else (t1' `TmApp` t2)
eval1 (TmVar x)                       = TmStop

-- special
eval1 TmLoop                          = TmLoop
eval1 TmStop                          = TmStop
eval1 (TmError e)                     = TmStop
eval1 t                               = TmError ("stuck term " ++ show t)


{- Evaluation methods -}

-- | Return the full computation of a term.
compute :: Term -> Computation
compute = safefulleval

-- | Step-by-step evaluation.
fulleval :: Term -> Computation
fulleval = takeWhile (/= TmStop) . (iterate eval1)

-- | Big-step evaluation.
eval :: Term -> Term
eval = last . fulleval

-- | Step-by-step evaluation, avoiding easy-to-catch loops.
safefulleval :: Term -> Computation
safefulleval = (protectWith TmLoop TmStop) . (iterate eval1)

-- | Big-step evaluation, avoiding easy-to-catch loops.
safeeval :: Term -> Term
safeeval = last . safefulleval

-- | Step-by-step smart and safe evaluation, where smart means that it can accept parsed terms.
smartfulleval :: Either ParseError Program -> String
smartfulleval parsed = case parsed of
    Left e       -> show e
    Right []     -> "Done.\n"
    Right (t:ts) -> (showMany . safefulleval) t ++ "\n\n" ++ (smartfulleval (Right ts))

-- | Big-step smart and safe evaluation, where smart means that it can accept parsed terms.
smarteval :: Either ParseError Program -> String
smarteval parsed = case parsed of
    Left e       -> show e
    Right terms  -> showMany $ map safeeval $ terms

