-- | Parser for types, typings and contexts.
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module ContextParser
    ( parseContext
    , typize
    , typingize
    , readtyping
    ) where

import TypeSystem
import qualified ProgramParser as PP
import qualified Text.ParserCombinators.Parsec.Token as T
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Error
import Text.ParserCombinators.Parsec.Language


{- M.V.P.s -}

-- | Turn a string into a typing.
typingize :: String -> Typing
typingize input = case (parse readtyping "(typing)" input) of
                  Left e       -> ("error", TyWrong (show e))
                  Right (x, t) -> (x, t)

-- | Turn a string into a type.
typize :: String -> Type
typize input = case (parse readtype "(typize)" input) of
               Left e  -> TyWrong (show e)
               Right t -> t

-- | Given a string (should be the contents of a context file),
-- try to parse it as a Context.
parseContext :: String -> Either ParseError Context
parseContext = parse contextFormat "(context file)"


{- High-level parsers -}

contextFormat = endBy contextLine (many1 (oneOf "\n\r"))
contextLine   = between (many $ oneOf "\n\r\t ") (many1 $ oneOf ";") readtyping


{- Language of types: definition -}

contextLang  = LanguageDef
               { commentStart = "/*"
               , commentEnd = "*/"
               , commentLine = "%"
               , nestedComments = False
               , identStart = upper
               , identLetter = letter <|> digit
               , opStart = letter
               , opLetter = letter <|> digit
               , reservedNames = [ "Bool"
                                 , "Nat"
                                 ]
               , reservedOpNames = ["->", "::"]
               , caseSensitive = True
               }

contextLexer = T.makeTokenParser contextLang


{- Handy shortcuts -}

whiteSpace  = T.whiteSpace contextLexer
symbol      = T.symbol contextLexer
parens      = T.parens contextLexer
identifier  = T.identifier contextLexer
reserved    = T.reserved contextLexer
reservedOp  = T.reservedOp contextLexer


{- Combinator parsers -}

-- | Parser for a full typing statement.
readtyping :: GenParser Char st Typing
readtyping = do x <- PP.identifier
                reservedOp "::"                
                t <- readtype
                return (x, t)

-- | Parser for a single type.
readtype :: GenParser Char st Type
readtype = try readbase
           <|> try readarrow
           <|> try readpar
           <|> try readcustom
           <?> "any term"


{- Helper parsers -}

readbase    = try readbool
              <|> try readnat
              <?> "base type"

readpar     = between (symbol "(") (symbol ")") readtype

readarrow   = between (symbol "(") (symbol ")") readarrow'

readarrow'  = do types <- sepBy readtype (reservedOp "->")
                 return $ arrowize types

readcustom  = do customName <- identifier
                 return (TyCustom customName)

readbool    = do reserved "Bool"
                 return TyBool

readnat     = do reserved "Nat"
                 return TyNat

