-- | Main tsoutapl module.
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module Main
    ( main
    , version
    ) where

import Sidekicks            ( showMany )
import Terms                ( Program )
import TypeSystem           ( Context, showTypedComputation )
import ProgramParser        ( parseProgram )
import ContextParser        ( parseContext )
import Calculus             ( compute, smartfulleval )
import System.Environment   ( getArgs )
import IO                   ( stderr, hPutStrLn )
import Text.Printf          ( printf, hPrintf )
import Text.ParserCombinators.Parsec.Error

version     = "1.2"


{- Messages -}

msgNoArgs   = "No arguments given.  But of course you _have_ read the README file, haven't you?"
msgRTFM     = "RTFM Error: Illegal arguments."
msgVersion  = "tsoutapl-" ++ version
msgAuthor   = "Thanos Tsouanas <thanos@sians.org>"
msgWebsite  = "http://mpla.math.uoa.gr/~thanos/code/tsoutapl/"


{- Main -}

main = do
    args <- getArgs
    case args of
        []            -> do putStrLn msgNoArgs
        ["-h"]        -> do putStrLn msgNoArgs
        ["-v"]        -> do putStrLn msgVersion
                            putStrLn msgAuthor
                            putStrLn msgWebsite
        ["-u", pf]    -> do sayf "*** Running tsoutapl-%s in untyped mode.\n" version
                            sayf "*** Reading program file %s... " pf
                            prgFile <- readFile pf
                            say "Okay."
                            putStrLn $ smartfulleval $ parseProgram prgFile
        ["-u", _, _]  -> do error "Untyped mode needs only a single (program) file."
        [pf]          -> do sayf "*** Running tsoutapl-%s in typed mode.\n" version
                            sayf "*** Reading program file %s... " pf
                            prgFile <- readFile pf
                            say "Okay."
                            putStrLn (let parsedContext = Right []
                                          parsedTerms   = parseProgram prgFile
                                      in smartfulltypedeval parsedContext parsedTerms)
        [pf, cf]      -> do sayf "*** Running tsoutapl-%s in untyped mode.\n" version
                            sayf "*** Reading program file %s... " pf
                            prgFile <- readFile pf
                            say "Okay."
                            sayf "*** Reading context file %s... " cf
                            ctxFile <- readFile cf
                            say "Okay."
                            putStrLn (let parsedContext = parseContext ctxFile
                                          parsedTerms   = parseProgram prgFile
                                      in smartfulltypedeval parsedContext parsedTerms)
        __________    -> error msgRTFM


{- Helpers -}

say  = hPutStrLn stderr
sayf = hPrintf stderr

-- | Given a (hopefully) context and a (hopefully) program
-- evaluate each term pretty-printing each step as you go along.
smartfulltypedeval :: Either ParseError Context -> Either ParseError Program -> String
smartfulltypedeval parsedContext parsedTerms = case (parsedContext, parsedTerms) of
    ( Left e1    , Left e2       ) -> error (showMany [e1, e2])
    ( Left e     , _             ) -> show e
    ( _          , Left e        ) -> show e
    ( Right _    , Right []      ) -> "Done.\n"
    ( Right ctx  , Right (t:ts)  ) -> showTypedComputation ctx (compute t)
                                      ++ "\n"
                                      ++ smartfulltypedeval (Right ctx) (Right ts)
