-- | Parser for terms and programs.
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module ProgramParser
    ( parseProgram
    , termize
    , identifier
    , readterm
    ) where

import Terms
import Numeric ( readDec )
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Error
import Text.ParserCombinators.Parsec.Language
import qualified Text.ParserCombinators.Parsec.Token as T


{- M.V.P.s -}

-- | Turn a string into a term.
termize :: String -> Term
termize input = case (parse readterm "(termize)" input) of
                Left e  -> TmError (show e)
                Right t -> t


-- | Given a string (should be the contents of a context file),
-- try to parse it as a Context.
parseProgram :: String -> Either ParseError [Term]
parseProgram = parse programFormat "(program file)"


{- High-level parsers -}

programFormat = endBy programLine (many1 (oneOf "\n\r"))
programLine   = between (many $ oneOf "\n\r\t ") (many1 $ oneOf ";") readterm


{- Language of terms: definition -}

programLang = LanguageDef
              { commentStart = "/*"
              , commentEnd = "*/"
              , commentLine = "%"
              , nestedComments = False
              , identStart = lower
              , identLetter = letter
              , opStart = letter
              , opLetter = letter <|> digit
              , reservedNames = [ "\\"
                                , "if"
                                , "then"
                                , "else"
                                , "true"
                                , "false"
                                , "succ"
                                , "pred"
                                , "iszero"
                                ]
              , reservedOpNames = [".", " "]
              , caseSensitive = True
              }

programLexer  = T.makeTokenParser programLang


{- Handy helpers -}

whiteSpace  = T.whiteSpace  programLexer
symbol      = T.symbol      programLexer
parens      = T.parens      programLexer
identifier  = T.identifier  programLexer
reserved    = T.reserved    programLexer
reservedOp  = T.reservedOp  programLexer


{- Combinator parsers -}

-- | Parser for a single term.
readterm :: GenParser Char st Term
readterm = try readnat
           <|> try readoper
           <|> try readabs
           <|> try readpar
           <|> try readapp
           <|> try readvar
           <|> try readbool
           <?> "any term"


{- Helper parsers -}

readoper    = try readif
              <|> try readsucc
              <|> try readpred
              <|> try readiszero
              <?> "operator"

readbool    = try readtrue
              <|> try readfalse
              <?> "boolean value"

readpar     = between (symbol "(") (symbol ")") readterm

readnat     = do natName <- many1 digit
                 return $ (natFromInt . fst . head . readDec) natName

readabs     = do string "\\"
                 boundVarName <- identifier
                 char '.'
                 skipMany $ char ' '
                 bodyTerm <- readterm
                 return (TmAbs boundVarName bodyTerm)

readapp     = do char '('
                 t1 <- readterm
                 whiteSpace
                 t2 <- readterm
                 char ')'
                 return (TmApp t1 t2)

readvar     = do varName <- identifier
                 return (TmVar varName)

readsucc    = do reserved "succ"
                 t <- readterm
                 return (TmSucc t)

readpred    = do reserved "pred"
                 t <- readterm
                 return (TmPred t)

readiszero  = do reserved "iszero"
                 t <- readterm
                 return (TmIsZero t)

readif      = do reserved "if"
                 skipMany $ char ' '
                 t <- readterm
                 skipMany $ char ' '
                 reserved "then"
                 skipMany $ char ' '
                 t1 <- readterm
                 skipMany $ char ' '
                 reserved "else"
                 skipMany $ char ' '
                 t2 <- readterm
                 return (TmIf t t1 t2)

readtrue    = do reserved "true"
                 return TmTrue

readfalse   = do reserved "false"
                 return TmFalse

