-- | Definition of terms.
--
-- Author:   Thanos Tsouanas <thanos@sians.org>
-- License:  BSD
module Terms
    ( Term(..)
    , Computation
    , Program
    , isNatValue
    , isBoolValue
    , isAbsValue
    , isValue
    , natFromInt
    , (|->)
    , (|@>)
    , (@=)
    ) where

import Sidekicks
import List ( union, (\\) )


-- | Terms.
data Term = TmTrue              -- ^ Boolean value true
          | TmFalse             -- ^ Boolean value false
          | TmIf Term Term Term -- ^ Branching
          | TmZero              -- ^ Natural number zero
          | TmSucc Term         -- ^ Successor
          | TmPred Term         -- ^ Predecessor
          | TmIsZero Term       -- ^ Zero test
          | TmVar String        -- ^ Variable
          | TmAbs String Term   -- ^ Lambda abstraction
          | Term `TmApp` Term   -- ^ Term application
          | TmStop              -- ^ Stop of evaluation meta-term
          | TmLoop              -- ^ A term that evaluates to itself without stopping
          | TmError String      -- ^ A term representing an error
          deriving (Eq)         -- Beware, this Eq never heard of the alpha-conversion.

instance Show Term where
    show TmTrue            = "true"
    show TmFalse           = "false"
    show (TmIf t t1 t2)    = "(if " ++ show t ++ " then " ++ show t1 ++ " else " ++ show t2 ++ ")" 
    show TmZero            = "0"
    show (TmSucc t)
        | isNatValue t     = showNumeral 1 t 
        | otherwise        = "(succ " ++ show t ++ ")"
        where showNumeral acc TmZero     = show acc
              showNumeral acc (TmSucc t) = showNumeral (acc + 1) t
    show (TmPred t)        = "(pred " ++ show t ++ ")"
    show (TmIsZero t)      = "(iszero " ++ show t ++ ")"
    show (TmVar x)         = "" ++ x ++ ""
    show (TmAbs x t)       = "(\\" ++ x ++ ". " ++ show t ++ ")"
    show (t1 `TmApp` t2)   = "(" ++ show t1 ++ " " ++ show t2 ++ ")"
    show TmStop            = "$"
    show TmLoop            = "@loop@"
    show (TmError e)       = "{Error: " ++ show e ++ "}"


-- | List of terms that represents a computation.
type Computation = [Term]


-- | List of terms that represent a program.
type Program = [Term]


-- | Turn an integral number to a (succ (succ (... zero)...) term.
natFromInt :: Integral t => t -> Term
natFromInt 0       = TmZero
natFromInt (n + 1) = TmSucc (natFromInt n)
natFromInt _       = TmError "Negative nat?!"


{- Checkers -}

-- | Check if the given term represents a numeral.
isNatValue :: Term -> Bool
isNatValue TmZero       = True
isNatValue (TmSucc t)   = isNatValue t
isNatValue _            = False

-- | Check if the given term is a Bool.
isBoolValue :: Term -> Bool
isBoolValue TmTrue      = True
isBoolValue TmFalse     = True
isBoolValue _           = False

-- | Check if the given term is a functional value.
isAbsValue :: Term -> Bool
isAbsValue (TmAbs _ t)  = isValue t
isAbsValue _            = False

-- | Check if the given term is a value for Nat or Bool.
isValue :: Term -> Bool
isValue                 = passAny valueCheckers
                          where valueCheckers = [isNatValue, isBoolValue, isAbsValue]


{- Variables -}

{- Helper functions -}

-- Return a list of all variables that appear in the given term.
variables :: Term -> [String]
variables term = case term of
    (TmVar x)       -> [x]
    (TmAbs x t)     -> variables t
    (t1 `TmApp` t2) -> foldl1 union $ map variables [t1, t2]
    (TmIf t t1 t2)  -> foldl1 union $ map variables [t, t1, t2]
    (TmSucc t)      -> variables t
    (TmPred t)      -> variables t
    (TmIsZero t)    -> variables t
    _______________ -> []

-- FV(t), sort of.
freeVariables :: Term -> [String]
freeVariables term = case term of
    (TmVar x)       -> [x]
    (TmAbs x t)     -> (freeVariables t) \\ [x]
    (t1 `TmApp` t2) -> foldl1 union $ map freeVariables [t1, t2]
    (TmIf t t1 t2)  -> foldl1 union $ map freeVariables [t, t1, t2]
    (TmSucc t)      -> freeVariables t
    (TmPred t)      -> freeVariables t
    (TmIsZero t)    -> freeVariables t
    _______________ -> []

-- Handy shortcut for x \in FV(t)
freeIn :: String -> Term -> Bool
x `freeIn` t = x `elem` (freeVariables t)

-- What, this is handy too!
notFreeIn :: String -> Term -> Bool
x `notFreeIn` t = not (x `freeIn` t)


{- Variables names -}

niceNames :: [String]
niceNames = ["x", "y", "z"]

-- BEWARE!  Infinite list: x, y, z, x', y', z', x'', y'', z'', ...
niceNamesPrimed :: [String]
niceNamesPrimed = zipWith (++) names decorators
    where names      = cycle niceNames
          decorators = [ replicate n '\'' | n <- [0..]]

-- BEWARE!  Infinite list: x0, y0, z0, x1, y1, z1, x2, y2, z2, ...
niceNamesIndexed :: [String]
niceNamesIndexed = zipWith (++) names decorators
    where names      = cycle niceNames
          decorators = map show [ n `div` (length niceNames) | n <- [0..] ]

-- | Something like a Hartogs operator for finding fresh variables.
hartogs :: [String] -> String
hartogs taken = head $ dropWhile (`elem` taken) niceNamesIndexed


{- Substitutions and all that Jazz -}

-- | Substitution in lambda-calculus terms.
(|->) :: String -> Term -> Term -> Term

(x |-> s) (TmVar y)   | x == y            = s
                      | otherwise         = (TmVar y)

(x |-> s) (TmAbs y t) | (x == y)          = TmAbs y t
                      | (y `notFreeIn` s) = TmAbs y ((x |-> s) t)
                      | otherwise         = let w       = hartogs badvars
                                                badvars = freeVariables s `union` variables t
                                            in (x |-> s) $ (y |@> w) (TmAbs y t)

(x |-> s) term = case term of
    (TmSucc t)      -> TmSucc t'       where t' = (x |-> s) t
    (TmPred t)      -> TmPred t'       where t' = (x |-> s) t
    (TmIsZero t)    -> TmIsZero t'     where t' = (x |-> s) t
    (t1 `TmApp` t2) -> t1' `TmApp` t2' where [t1', t2']     = map (x |-> s) [t1, t2]
    (TmIf t t1 t2)  -> TmIf t' t1' t2' where [t', t1', t2'] = map (x |-> s) [t, t1, t2]
    t               -> t


-- | Renaming in the spirit of the alpha-conversion of lambda-calculus.
-- BEWARE!  This isn't "safe": 'w' must be /correctly/ picked before-hand.
-- In fact, this is the most brain-dead substitution implementation.
(|@>) :: String -> String -> Term -> Term
(x |@> w) term = case term of
    (TmSucc t)      -> TmSucc t'         where t' = (x |@> w) t
    (TmPred t)      -> TmPred t'         where t' = (x |@> w) t
    (TmIsZero t)    -> TmIsZero t'       where t' = (x |@> w) t
    (t1 `TmApp` t2) -> (t1' `TmApp` t2') where [t1', t2']     = map (x |@> w) [t1, t2]
    (TmIf t t1 t2)  -> TmIf t' t1' t2'   where [t', t1', t2'] = map (x |@> w) [t, t1, t2]
    (TmVar y)       -> if (x == y) then (TmVar w) else (TmVar y)
    (TmAbs y t)     -> if (x == y) then (TmAbs w t') else (TmAbs y t') where t' = (x |@> w) t
    _______________ -> term


-- | alpha-conversion equality.
(@=) :: Term -> Term -> Bool
t1 @= t2 = case (t1, t2) of 
    ( (TmTrue)        , (TmTrue)        ) -> True
    ( (TmFalse)       , (TmFalse)       ) -> True
    ( (TmVar x)       , (TmVar y)       ) -> x == y
    ( (t1 `TmApp` t2) , (s1 `TmApp` s2) ) -> t1 @= s1 && t2 @= s2
    ( (TmSucc t)      , (TmSucc s)      ) -> t @= s
    ( (TmPred t)      , (TmPred s)      ) -> t @= s
    ( (TmIsZero t)    , (TmPred s)      ) -> t @= s
    ( (TmAbs x t)     , (TmAbs y s)     ) -> t @= (y |@> x) s
    _____________________________________ -> False


